/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include "Editors/GuitarEditor.h"
#include "Midi/Sequence.h"
#include "Midi/Track.h"
#include "Pickers/VolumeSlider.h"
#include "Pickers/TuningPicker.h"
#include "Images/ImageProvider.h"
#include "GUI/GLPane.h"

#include "OpenGL.h"

#include <string>

#include "Editors/RelativeXCoord.h"

namespace AriaMaestosa {
	
const int first_string_position = 17;
	
GuitarEditor::GuitarEditor(Track* track) : Editor(track)
{
    
    dragging_on_track=false;
    
    clickedOnNote=false;
    lastClickedNote=-1;
    
    tuningPicker = new TuningPicker(this);
    tuningPicker->loadTuning(1); // standard
    
}

/*
 * This is called when user changes the tuning. This tells the editor to change notes so that they match the new settings.
 */

void GuitarEditor::tuningUpdated()
{
    track->updateNoteTuning(ALL_NOTES);
}

void GuitarEditor::render(int dragX_arg, int mousey_current, int XBeforeDrag_arg, int mousey_initial, bool focus)
{
    
    if(!imagesLoaded()) return;
    
    const int string_amount = tuning.size();
    
    // prepare coords
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
    
    glEnable(GL_SCISSOR_TEST);
    // glScissor doesn't seem to follow the coordinate system so this ends up in all kinds of weird code to map to my coord system (from_y going down)
    glScissor(10, sequence->glPane-> getHeight() - (20+height + from_y+barHeight+20), width-15, 20+height);
    
    // white background
    glDisable(GL_TEXTURE_2D);
    
    glColor3f(1,1,1);
    glBegin(GL_QUADS);
    
    glVertex2f( 0, getEditorYStart());
    glVertex2f( 0, getYEnd());
    glVertex2f( getXEnd(), getYEnd());
    glVertex2f( getXEnd(), getEditorYStart());
    
    glEnd();
    
    drawVerticalMeasureLines(getEditorYStart()+first_string_position, getEditorYStart()+first_string_position+string_amount*10-10);

    // ------------------------------- draw strings -------------------------------
    glColor3f(0,0,0);
    glBegin(GL_LINES);
    
    for(unsigned int n=0; n<tuning.size(); n++)
	{
        glVertex2f( getEditorXStart(), getEditorYStart()+first_string_position+n*10 );
        glVertex2f( getXEnd(), getEditorYStart()+first_string_position+n*10 );
    }
    
    glEnd();

    
    // ---------------------- draw notes ----------------------------
    const int noteAmount = track->getNoteAmount();
    for(int n=0; n<noteAmount; n++)
	{
        int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
        int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
        
        // don't draw notes that won't visible
        if(x2<0) continue;
        if(x1>width) break;
        
        int string=track->getNoteString(n);
        int fret=track->getNoteFret(n);
        
        float volume=track->getNoteVolume(n)/127.0;
        
        if(track->isNoteSelected(n) and focus) glColor3f((1-volume)*1, (1-(volume/2))*1, 0);
        else glColor3f((1-volume)*0.9, (1-volume)*0.9, (1-volume)*0.9);
        
		// note body
        glBegin(GL_QUADS);
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+2);
        glVertex2f(x2-1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+2);
        glVertex2f(x2-1+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        glEnd();
		
		// lines around note
		glColor3f(0, 0, 0);
        glBegin(GL_LINES);
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+2);
        
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+3);
        glVertex2f(x2-1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+3);
        
        glVertex2f(x2+getEditorXStart(), getEditorYStart()+first_string_position+string*10+2);
        glVertex2f(x2+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        
        glVertex2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        glVertex2f(x2-1+getEditorXStart(), getEditorYStart()+first_string_position+string*10-2);
        glEnd();
        
        // fret number
        if(track->isNoteSelected(n)  and focus) glColor3f((1-volume)*1, (1-(volume/2))*1, 0);
        else glColor3f((1-volume)*0.9, (1-volume)*0.9, (1-volume)*0.9);
        
        glBegin(GL_QUADS);
        glVertex2f(x1+getEditorXStart()-1, getEditorYStart()+first_string_position+string*10-8);
        glVertex2f(x1+getEditorXStart()-1, getEditorYStart()+first_string_position+string*10);
        glVertex2f(x1+14+getEditorXStart(), getEditorYStart()+first_string_position+string*10);
        glVertex2f(x1+9+getEditorXStart(), getEditorYStart()+first_string_position+string*10-8);
        glEnd();
        
        if((!track->isNoteSelected(n) or !focus) && volume>0.5) glColor3f(1, 1, 1); // if note color is too dark, draw the fret number in whites
        else glColor3f(0, 0, 0);
        
        // convert the fret nubmer to chars, and draw it
        glRasterPos2f(x1+getEditorXStart(), getEditorYStart()+first_string_position+string*10+1);
        char buffer[2];
        sprintf (buffer, "%d", fret);
        
        for(int i=0; buffer[i]; i++)
		{
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, buffer[i]);
        }
        
    }//next
    
    
    // mouse drag
    if(!clickedOnNote)
	{
        
        if(dragging_on_track)
		{
            
            // -------------------- selection ----------------------
			if(selecting)
			{

                glColor3f(0,0,0);
                glBegin(GL_LINES);
                
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_current);
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_initial);
                
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_initial);
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_initial);
                
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_initial);
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_current);
                
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_current);
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_current);
                
                glEnd();
                
            }
			// ----------------------- add note (preview) --------------------
			else if(mousey_current<getYEnd()-15)
			{
                
                glColor3f(1, 0.85, 0);
                
                const int preview_x1=
                    (int)(
                          (snapMidiTickToGrid(mousex_initial.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks()) -
                           sequence->getXScrollInMidiTicks())*sequence->getZoom()
                          );
                const int preview_x2=
                    (int)(
                          (snapMidiTickToGrid(mousex_current.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks()) -
                           sequence->getXScrollInMidiTicks())*sequence->getZoom()
                          );
                
                int string = (int)round((mousey_initial - getEditorYStart() - first_string_position)/10.0);
                
                if(!(preview_x1<0 || preview_x2<0 || string<0 || (unsigned int)string>tuning.size()-1))
				{
                    glBegin(GL_QUADS);
                    
                    glVertex2f(preview_x1+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5);
                    glVertex2f(preview_x1+getEditorXStart(), string*10+10 + getEditorYStart() + first_string_position - 5);
                    glVertex2f(preview_x2+getEditorXStart(), string*10+10 + getEditorYStart() + first_string_position - 5);
                    glVertex2f(preview_x2+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5);
                    
                    glEnd();
                }
                
            }// end if selection or addition
            
        }// end if dragging on track
        
    }// if !clickedOnNote
     // ------------------------- move note (preview) -----------------------
    if(clickedOnNote)
	{
        
        glDisable(GL_TEXTURE_2D);
        
        glColor4f(1, 0.85, 0, 0.5);
        
        int xstep=mousex_current.getRelativeTo(MIDI)-mousex_initial.getRelativeTo(MIDI);
        int ystep=mousey_current-mousey_initial;
        
        xstep = (int)( snapMidiTickToGrid(xstep)*sequence->getZoom() );
        ystep=(int)round(ystep/10.0);
        
        // move a single note
        if(lastClickedNote!=-1)
		{
            const int x1=track->getNoteStartInPixels(lastClickedNote) - sequence->getXScrollInPixels();
            const int x2=track->getNoteEndInPixels(lastClickedNote) - sequence->getXScrollInPixels();
            const int string=track->getNoteString(lastClickedNote);

            glBegin(GL_QUADS);
            glVertex2f(x1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + ystep*10);
            glVertex2f(x1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + 10 + ystep*10);
            glVertex2f(x2-1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 +10 + ystep*10);
            glVertex2f(x2-1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + ystep*10);
            glEnd();  
        }
		else
		{
            // move a bunch of notes
            
            for(int n=0; n<track->getNoteAmount(); n++){
                if(!track->isNoteSelected(n)) continue;
                
                const int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
                const int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
                const int string=track->getNoteString(n);
                
                glBegin(GL_QUADS);
                glVertex2f(x1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + ystep*10);
                glVertex2f(x1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + 10 + ystep*10);
                glVertex2f(x2-1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 +10 + ystep*10);
                glVertex2f(x2-1+xstep+getEditorXStart(), string*10 + getEditorYStart() + first_string_position - 5 + ystep*10);
                glEnd();
            }//next
            
        }//end if lastClickedNote!=-1
        
    }//end if clickedOnNote
    
    // -----------------------------------------------------------------
    // left part with string names
    // -----------------------------------------------------------------
    glLoadIdentity();
    
    // grey background
    if(!focus) glColor3f(0.4, 0.4, 0.4);
    else glColor3f(0.8, 0.8, 0.8);
    glBegin(GL_QUADS);
    
    glVertex2f( 0, getEditorYStart());
    glVertex2f( 0, getYEnd());
    glVertex2f( getEditorXStart()-3, getYEnd());
    glVertex2f( getEditorXStart()-3, getEditorYStart());
    
    glEnd();
    
    // string names
    glColor3f(0,0,0);
    
    for(int n=0; n<string_amount; n++)
	{
        glRasterPos2f(getEditorXStart()-25, getEditorYStart()+20+n*10);
        
        const int octave=tuning[n]/12;
        const int note=tuning[n]%12;
        
        switch(note){
            case 0:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'B');
                break;
            case 1:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'A');
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, '#');
                break;
            case 2:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'A');
                break;
            case 3:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'G');
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, '#');
                break;
            case 4:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'G');
                break;
            case 5:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'F');
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, '#');
                break;
            case 6:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'F');
                break;
            case 7:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'E');
                break;
            case 8:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'D');
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, '#');
                break;
            case 9:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'D');
                break;
            case 10:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'C');
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, '#');
                break;
            case 11:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, 'C');
                break;
        } // end switch
        
        char buffer[2];
        sprintf (buffer, "%d", 10-octave);
        
        for(int i=0; buffer[i]; i++)
		{
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, buffer[i]);
        }
        
    }//next
    
    
    
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_SCISSOR_TEST);
    
}

void GuitarEditor::mouseExited(int dragX_arg, int mousey_current, int XBeforeDrag_arg,int mousey_initial)
{
	RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
	
	selectNotesInRect(mousex_current, mousey_current, mousex_initial, mousey_initial);
	selecting = false;
	dragging_on_track = false;
	
	sequence->glPane->render();
}

void GuitarEditor::mouseDown(const int x_arg, const int y)
{

	selecting = false;
	
    // prepare coords
    RelativeXCoord<int> x(sequence);
    x.setValue(x_arg, WINDOW);
    
    // find if click is in track
    if(y<getYEnd()-5 and y>getEditorYStart())
	{
        dragging_on_track=true;
        
        // user clicked on left bar to change tuning
        if(x.getRelativeTo(EDITOR)<0 and x.getRelativeTo(EDITOR)>-50)
		{
            sequence->glPane->PopupMenu(tuningPicker,x.getRelativeTo(WINDOW),y);
        }
        
    }
    
    // find if user clicked on a note
	if(dragging_on_track)
	{
		
		clickedOnNote=false;
		const int noteAmount = track->getNoteAmount();
		// iterate through note n reverse order (last drawn note appears on top and must be first selected)
		for(int n=noteAmount-1; n>-1; n--)
		{
			const int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
			const int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
			
			const int string=track->getNoteString(n);
			
			if(x.getRelativeTo(EDITOR) > x1 and
			   x.getRelativeTo(EDITOR) < x2 and
			   y < getEditorYStart()+first_string_position+string*10+4 and
			   y > getEditorYStart()+first_string_position+string*10-5)
			{
				
				clickedOnNote=true;
				
				if(track->isNoteSelected(n) and !sequence->glPane->isSelectLessPressed())
				{
					// clicked on a selected note
					lastClickedNote=-1;
				}
				else lastClickedNote=n;
				
				track->playNote(n);
				
				break;
			}
		}//next
		
	}
    
    // user clicked on a note. select it.
    if(clickedOnNote && lastClickedNote > -1)
	{
        track->selectNote(ALL_NOTES, false);
        track->selectNote(lastClickedNote, true);
    }
    
}

void GuitarEditor::mouseDrag(const int mousex_current, const int mousey_current, const int mousex_initial, const int mousey_initial)
{

	// if click is not on track anymore, terminate selection
    if(!(mousey_current<getYEnd()-5 and mousey_current>getEditorYStart()))
	{
        mouseExited(mousex_current, mousey_current, mousex_initial, mousey_initial);
	}
	
		
    if( !clickedOnNote and abs(mousey_current-mousey_initial)>10 ) selecting = true;
}

void GuitarEditor::mouseUp(const int dragX_arg, const int mousey_current, const int XBeforeDrag_arg, const int mousey_initial)
{
    // prepare coords
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
    
    if(!clickedOnNote)
	{
		
		if(dragging_on_track)
		{
			
			// -------------------- selection -----------------
			if(selecting)
			{
				selectNotesInRect(mousex_current, mousey_current, mousex_initial, mousey_initial);
			}
			// ----------------------- add note --------------------
			else
			{
				int snapped_x1=snapMidiTickToGrid(mousex_initial.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks());
				int snapped_x2=snapMidiTickToGrid(mousex_current.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks());
				
				int string = (int)round((mousey_initial - getEditorYStart() - first_string_position)/10.0);
				
				if(string<0) return; //invalid note, don't add it
				if(string>(int)tuning.size()) return; //invalid note, don't add it
				
				track->addNote(tuning[string], snapped_x1, snapped_x2, 80, string );
				
			}// end if selection or addition
		}// end if dragging on track
		
    }
	else
	{
        // ------------ move note ---------------
		
		int xstep=mousex_current.getRelativeTo(MIDI) - mousex_initial.getRelativeTo(MIDI);
		int ystep=mousey_current - mousey_initial;
		
		xstep = snapMidiTickToGrid(xstep);
		ystep=(int)round(ystep/10.0);
		
		if(xstep == 0 and ystep == 0)
		{
			//if(lastClickedNote!=-1) track->playNote(lastClickedNote);
			goto end_of_func;
		}
		
		if(lastClickedNote!=-1)
		{
			// move a single note
			track->moveNotes(xstep, ystep, lastClickedNote);
		}
		else
		{
			// move selected notes
			track->moveNotes(xstep, ystep, SELECTED_NOTES);     
		}
		
    }
    
end_of_func:
	
    dragging_on_track=false;
    clickedOnNote=false;
     
}

void GuitarEditor::selectNotesInRect(RelativeXCoord<int>& mousex_current, int mousey_current, RelativeXCoord<int>& mousex_initial, int mousey_initial)
{
	for(int n=0; n<track->getNoteAmount(); n++)
	{
		const int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels(); // on-screen pixel where note starts
		const int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels(); // on-screen pixel where note ends
		
		const int string=track->getNoteString(n);
		
		// check if note is within selection boundaries
		if(x1 > std::min(mousex_current.getRelativeTo(EDITOR), mousex_initial.getRelativeTo(EDITOR)) and
		   x2 < std::max(mousex_current.getRelativeTo(EDITOR), mousex_initial.getRelativeTo(EDITOR)) and
		   std::min(mousey_current, mousey_initial) < getEditorYStart()+first_string_position+string*10 and
		   std::max(mousey_current, mousey_initial) > getEditorYStart()+first_string_position+string*10)
		{
			track->selectNote(n, true);
		}
		else
		{
			track->selectNote(n, false);
		}
	}//next
}

void GuitarEditor::rightClick(const int x, const int y)
{

	// find on which note the user clicked
	const int noteAmount = track->getNoteAmount();
	// iterate through note n reverse order (last drawn note appears on top and must be first selected)
	for(int n=noteAmount-1; n>-1; n--)
	{
        int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
        int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
        
        int string=track->getNoteString(n);
        //int fret=track->getNoteFret(n);
        
        if(x > x1+getEditorXStart() and
           x < x2+getEditorXStart() and
           y < getEditorYStart()+first_string_position+string*10+2 and
           y > getEditorYStart()+first_string_position+string*10-5)
		{
            
            wxPoint winCoord=sequence->glPane->ClientToScreen(wxPoint(x,y));
            volumeSlider->show(winCoord.x, winCoord.y, n, track);
            
            return;
        }
    }//next
    
    
}

int GuitarEditor::getYScrollInPixels()
{
    /*Unused*/
    return -1;
}

GuitarEditor::~GuitarEditor()
{
    delete tuningPicker;
}


// rearrange notes smartly (chooses on which string each note should go)
void GuitarEditor::rearrangeNotes()
{
	
sequence->importing = true; // just to make sure notes are not played while reordering
	
	std::vector<int> candidates;
	
    const int noteAmount = track->getNoteAmount();
    for(int n=0; n<noteAmount; n++)
	{
		
		// find notes that have the same tick
		candidates.clear();
        int x1=track->getNoteStartInPixels(n);
		candidates.push_back(n);
		
		// found 2 consecutive notes with same tick - check if there are more
		if(n<noteAmount-1 and track->getNoteStartInPixels(n+1)==x1)
		{
		    n++;
		    while(n<noteAmount and track->getNoteStartInPixels(n)==x1)
			{
#ifdef _MORE_DEBUG_CHECKS
				static int iters=0; iters++; if(iters>10000000){ std::cout << "infinite loop" << std::endl; exit(1);}
#endif
				candidates.push_back(n);
				n++;
			}
			n--;
		}
		
		if(candidates.size()>1)
		{ // we found many notes with the same tick

			const int size = candidates.size();
			
			// check if any of them are selected
			bool selected = false;
			for(int i=0; i<size; i++)
			{
				if(track->isNoteSelected( candidates[i] )) selected=true;
			}
			if(!selected) continue;
			
			// check if more than one are on the same string (if they overlap)
			bool need_to_reorder=false;
			
			for(int a=0; a<size; a++)
			{
				for(int b=0; b<size; b++)
				{
					
					if(a != b) continue;
					
					if(track->getNoteString(candidates[a])==track->getNoteString(candidates[b]) /*overlapping notes*/
					   or  /*split powerchord*/
					   (abs(track->getNotePitchID(candidates[b])-track->getNotePitchID(candidates[b]))==7
						and abs(track->getNoteString(candidates[a])-track->getNoteString(candidates[b]))>1)
					   )
					{
						
						need_to_reorder = true;
						
					}//endif
					
				}//next
			}//next

			if(need_to_reorder)
			{
				
				// order in pitch order (lowest to highest)
				for(int i=0; i<size-1; i++)
				{
					
					if(track->getNotePitchID(candidates[i]) < track->getNotePitchID(candidates[i+1]))
					{
						
						int tmp = candidates[i+1];
						
						candidates[i+1] = candidates[i];
						candidates[i] = tmp;
						
						i=-1;
					}
					
				}// next
				
				
				// shift first note down if possible (if necessary)
				//std::cout << track->getNoteString(candidates[0]) << " - " << (size-1) << " = " << track->getNoteString(candidates[0])-size << std::endl;
				const int fstr = track->getNoteString(candidates[0])-(size-1); // calculate where highest note would end up this way
				if( fstr < 0)
				{ // if highest note wouldn't fit in available space, shift down all notes as many times as needed
					for(int n=0; n<abs(fstr); n++) track->shiftString( 1, candidates[0] );
				}
				
				// lay them out
				for(int i=1; i<size; i++)
				{
					// place lowest note on lowest needed string, then second lowest note on second lowest string, etc
					track->shiftString( track->getNoteString(candidates[0])-i-track->getNoteString(candidates[i]) , candidates[i]);
				}
			}
			
		}//endif
	}//next

sequence->importing = false;

}

}
